/GlobalSection(SolutionConfigurationPlatforms) = preSolution/ {
	n
	h
:loop1
	n
	/EndGlobalSection/ !{
		H
		b loop1
	}
	x
	s/Win32/x64/g
	a \	EndGlobalSection\r
}

/GlobalSection(ProjectConfigurationPlatforms) = postSolution/ {
	n
	h
:loop2
	n
	/EndGlobalSection/ !{
		H
		b loop2
	}
	x
	s/Win32/x64/g
	a \	EndGlobalSection\r
}

